/* platform_win.h
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   header file with platform specific definitions
 *
 * Copyright (C) 2003      <oskar>
 *
 *  2004-03-18 cft: extracted from platform-h
 *  2003-11-18 Oskar: added Windows stuff
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */

#ifndef PLATFORM_WIN_H
#define PLATFORM_WIN_H

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>

typedef unsigned char __u8;
typedef __u8 u_int8_t;
typedef __u8 uint8_t;
typedef unsigned short __u16;
typedef __u16 u_int16_t;
typedef __u16 uint16_t;
typedef unsigned int u32;
typedef unsigned int __u32;
typedef __u32 u_int32_t;
typedef __u32 uint32_t;
typedef unsigned __int64 __u64;
typedef __u64 uint64_t;
typedef long off_t;
typedef struct { volatile int counter; } atomic_t;
typedef unsigned short sa_family_t;
typedef unsigned short __kernel_mode_t;
typedef __kernel_mode_t mode_t;
typedef int (get_info_t)(char *, char **, off_t, int);

#define const64(x) (x##i64)
#define const64u(x) (x##Ui64)

__inline void set_bit(int nr, volatile void *addr) {
  /* TODO: Verify correct operation of this function */
  /* Can of course also be made much more effective
     by using inline assembler, something like:
     __asm {btsl nr, addr} */
  int divid = nr / (8 * sizeof(BYTE));
  int remain = nr % (8 * sizeof(BYTE));

  ((BYTE*)addr)[divid] |= (1 << remain);
}

__inline const __u16 __fswab16(__u16 x) {
  return (x << 8) | ((x >> 8) & 0x0ff);
}

__inline const __u32 __fswab32(__u32 x) {
  return
    (((__u32)x & (__u32)0x000000ffUL) << 24) |
    (((__u32)x & (__u32)0x0000ff00UL) <<  8) |
    (((__u32)x & (__u32)0x00ff0000UL) >>  8) |
    (((__u32)x & (__u32)0xff000000UL) >> 24) ;
}

__inline const __u64 __fswab64(__u64 x) {
  return
    (((__u64)x & (__u64)0x00000000000000ffUi64) << 56) |
    (((__u64)x & (__u64)0x000000000000ff00Ui64) << 40) |
    (((__u64)x & (__u64)0x0000000000ff0000Ui64) << 24) |
    (((__u64)x & (__u64)0x00000000ff000000Ui64) <<  8) |
    (((__u64)x & (__u64)0x000000ff00000000Ui64) >>  8) |
    (((__u64)x & (__u64)0x0000ff0000000000Ui64) >> 24) |
    (((__u64)x & (__u64)0x00ff000000000000Ui64) >> 40) |
    (((__u64)x & (__u64)0xff00000000000000Ui64) >> 56) ;
}

#define MODULE_AUTHOR(x) static const char module_author[] = x
#define MODULE_DESCRIPTION(x) static const char module_description[] = x
#define MODULE_LICENSE(x) static const char module_license[] = x
#define MODULE_PARM(dbgl, x) static const char module_parm_##dbgl[] = x
#define MODULE_PARM_DESC(dbgl, x) static const char module_parm_desc_##dbgl[] = x
#define EXPORT_SYMBOL(x) 
#define module_init(x) 
#define module_exit(x) 
#define MOD_INC_USE_COUNT DbgPrint("MOD_INC_USE_COUNT")
#define MOD_DEC_USE_COUNT DbgPrint("MOD_DEC_USE_COUNT")
#define SET_MODULE_OWNER(x) DbgPrint("SET_MODULE_OWNER")

/* These, i.e. Hz of computer and number of cycles
   since startup, are chosen to fit with our timer
   implementations; Now, when they are used, they
   make time values appear in milliseconds instead
   of cycles */
/* TODO: Perhaps use microseconds instead since this
   is what is used as input to the netbox_newtimer
   function */
#define HZ 1000
#define jiffies get_millis() /* unsigned long volatile */

/* These are just dummy values since they are not used
   for anything in the Windows version */
#define GFP_KERNEL 0
#define GFP_ATOMIC 1

#define IFNAMSIZ 16
/* Largest hardware address length */
#define MAX_ADDR_LEN 8
/* Octets (bytes) in one ethernet addr */
#define ETH_ALEN 6
/* Total octets (bytes) in header */
#define ETH_HLEN 14
/* Address Resolution packet */
#define ETH_P_ARP 0x0806
/* Internet Protocol packet */
#define ETH_P_IP 0x0800
/* Notifier values */
#define NOTIFY_DONE 0x0000 /* Don't care */
/* netdevice notifier chain */
#define NETDEV_UP 0x0001
#define NETDEV_DOWN 0x0002
#define NETDEV_REGISTER 0x0005
#define NETDEV_UNREGISTER 0x0006
/* Error numbers */
#define ENXIO 6
#define ENOMEM 12

/* TODO: Should be moved to central location */
#define PARAM_LENGTH 4
typedef unsigned char BYTE;
typedef unsigned long u_long;

/* Our own definitions of packet types */
#define PT_UNKNOWN		0
#define PT_IP			1
#define PT_ARP			2
#define PT_SAPF			3

/* Supported address families */
#define AF_UNSPEC               0

#define __swab16(x) __fswab16(x)
#define __swab32(x) __fswab32(x)
#define __swab64(x) __fswab64(x)

#define __cpu_to_be16(x) __swab16((x))
#define __be16_to_cpu(x) __swab16((x))
#define __cpu_to_be32(x) __swab32((x))
#define __be32_to_cpu(x) __swab32((x))
#define __cpu_to_be64(x) __swab64((x))
#define __be64_to_cpu(x) __swab64((x))

#define cpu_to_be64 __cpu_to_be64
#define be64_to_cpu __be64_to_cpu


#define __htons(x) __cpu_to_be16(x)
#define __ntohs(x) __be16_to_cpu(x)
#define htons(x) __htons(x)
#define ntohs(x) __ntohs(x)
#define ntohl(x) __be32_to_cpu(x)
#define htonl(x) __cpu_to_be32(x)

#define time_after(a,b) ((long)(b) - (long)(a) < 0)

#define DEBUG(l,s) {if((l)<=debuglevel){s;}}
static int debuglevel = 9;
#define printk DbgPrint

/* 10Mb/s ethernet header */
struct ether_header
{
  u_int8_t  ether_dhost[ETH_ALEN]; /* destination eth addr */
  u_int8_t  ether_shost[ETH_ALEN]; /* source ether addr	   */
  u_int16_t ether_type;		   /* packet type ID field */
};

/* ARP protocol opcodes */
#define	ARPOP_REQUEST  0x0001 /* ARP request  */
#define	ARPOP_REPLY    0x0002 /* ARP reply    */
#define	ARPOP_RREQUEST 0x0003 /* RARP request */
#define	ARPOP_RREPLY   0x0004 /* RARP reply   */

/* Ethernet arp pkt */
struct arp_pkt
{
  unsigned short ar_hrd;	   /* format of hardware address */
  unsigned short ar_pro;	   /* format of protocol address */
  unsigned char	 ar_hln;	   /* length of hardware address */
  unsigned char	 ar_pln;	   /* length of protocol address */
  unsigned short ar_op;		   /* ARP opcode (command)	 */
  unsigned char	 ar_sha[ETH_ALEN]; /* sender hardware address	 */
  unsigned char	 ar_sip[4];	   /* sender IP address		 */
  unsigned char	 ar_tha[ETH_ALEN]; /* target hardware address	 */
  unsigned char	 ar_tip[4];	   /* target IP address	         */
};

struct iphdr {
  /* Note: Little endian */
  __u8 ihl:4,
    version:4;
  __u8  tos;
  __u16 tot_len;
  __u16 id;
  __u16 frag_off;
  __u8  ttl;
  __u8  protocol;
  __u16 check;
  __u32 saddr;
  __u32 daddr;
  /* The options start here */
};

struct udphdr {
  __u16 source;
  __u16 dest;
  __u16 len;
  __u16 check;
};

struct notifier_block {
  int (*notifier_call)(struct notifier_block *self, unsigned long, void*);
  struct notifier_block *next;
  int priority;
};

struct list_head {
  struct list_head *next, *prev;
};

struct timer_list {
  struct list_head list;
  unsigned long expires;
  unsigned long data;
  void (*function)(unsigned long);

  /* Needed by the Windows version to keep
     track of which timer is which */
  PNDIS_MINIPORT_TIMER timer;
};

struct in_addr {
  __u32 s_addr;
};

struct sockaddr {
  sa_family_t sa_family; /* address family, AF_xxx */
  char sa_data[14];      /* 14 bytes of protocol address */
};

/* This struct is very slimmed and just used to fit
   the Windoze port */
struct in_ifaddr {
  u32 ifa_local;
};

/* This struct is very slimmed and just used to fit
   the Windoze port */
struct in_device {
  /* IP ifaddr chain */
  struct in_ifaddr *ifa_list;
};

/* This struct is very slimmed and just used to fit
   the Windoze port */
struct net_device_stats {
  unsigned long rx_packets; /* total packets received */
  unsigned long tx_packets; /* total packets transmitted */
  unsigned long rx_bytes;
  unsigned long tx_bytes;
};

/* This struct is very slimmed and just used to fit
   the Windoze port */
struct iw_statistics {
  /* Status - device dependent for now */
  __u16 status;
};

/* Socket buffer - holds e.g. an IP datagram or
   ARP packet; This struct is very slimmed and
   just used to fit the Windoze port */
struct sk_buff {
  /* Network layer header, i.e. pointer to start
     of contained IP or ARP packet */
  union {
    unsigned char *raw;
  } nh;

  /* Control buffer; free to use for every layer */
  char cb[48];

  /* Length of actual data, i.e. of the IP or
     ARP packet contained in the ethernet packet */
  unsigned int len;
  /* Data head pointer, i.e. pointer to the start
     of the ethernet packet containing the IP or
     ARP packet */
  unsigned char *data;
};

/* This struct is very slimmed and just used to fit
   the Windoze port */
struct net_device {
  /* The name of the interface */
  char name[IFNAMSIZ];

  /* Note: In the Windows version currently
     only one device is handled */
  struct net_device *next;

  /* Interface index. Unique device identifier */
  int ifindex;

  struct net_device_stats* (*get_stats)(struct net_device *dev);
  struct iw_statistics* (*get_wireless_stats)(struct net_device *dev);

  /* Interface MTU value */
  unsigned mtu;

  /* hardware hdr length */
  unsigned short hard_header_len;

  /* pointer to private data */
  void *priv;

  /* Interface address info */
  /* hw address */
  unsigned char dev_addr[MAX_ADDR_LEN];

  /* Protocol specific pointers */
  /* IPv4 specific data */
  void *ip_ptr;
  /* Max frames per queue allowed */
  unsigned long tx_queue_len;

  /* Number of references to this device */
  atomic_t refcnt;

  /* Pointers to interface service routines */
  int (*open)(struct net_device *dev);
  int (*stop)(struct net_device *dev);
  int (*hard_start_xmit) (struct sk_buff *skb,
			  struct net_device *dev);
  int (*hard_header) (struct sk_buff *skb,
		      struct net_device *dev,
		      unsigned short type,
		      void *daddr,
		      void *saddr,
		      unsigned len);
};

struct packet_type {
  /* This is really htons(ether_type) */
  unsigned short type;
  /* NULL is wildcarded here */
  struct net_device *dev;
  int (*func) (struct sk_buff*, struct net_device*, struct packet_type*);
  /* Private to the packet type */
  void *data;
  struct packet_type *next;
};

/* This struct is very slimmed and just used to fit
   the Windoze port */
struct proc_dir_entry {
  unsigned short low_ino;
  unsigned short namelen;
  /* ... */
};

/* TODO: Perhaps implement this more correctly
   later */
struct rwlock_t {
  volatile unsigned int lock;
};

struct rwlock_t dev_base_lock;
/* Note: We set this to NULL in the Windows
   version in order to avoid problems */
static struct net_device *dev_base = NULL;

/* Some NDIS stuff */
#define NDIS_DIR_RECEIVE 0
#define NDIS_DIR_SEND 1
/* Structure used for modeling queued (ethernet)
   packets */
typedef struct _QUEUED_PACKET
{
  PVOID buffer;
  int length;
  int direction;
  
  /* Contains list entry to be able to include
     it in a queue */
  LIST_ENTRY listEntry;
} QUEUED_PACKET, *PQUEUED_PACKET;

/* Protocol reserved part of the packet */
typedef struct _PROTOCOL_RESERVED
{
  PNDIS_PACKET OriginalPkt;
} RSVD, *PRSVD;

void* kmalloc(unsigned int size, int flags);
void kfree(const void *objp);
void kfree_skb(struct sk_buff *skb);
void init_timer(struct timer_list *timer);
void add_timer(struct timer_list *timer);
int del_timer(struct timer_list *timer);
struct net_device* dev_get_by_index(int ifindex);
struct net_device* dev_get_by_name(const char *name);
void dev_put(struct net_device *dev);
struct net_device* init_etherdev(struct net_device *dev, int sizeof_priv);
void unregister_netdev(struct net_device *dev);
int register_netdevice_notifier(struct notifier_block *nb);
int unregister_netdevice_notifier(struct notifier_block *nb);
void netif_start_queue(struct net_device *dev);
void netif_stop_queue(struct net_device *dev);
void get_random_bytes(void *buf, int nbytes);
void read_lock(struct rwlock_t *rw);
void read_unlock(struct rwlock_t *lock);
unsigned char getRandomByte();
struct proc_dir_entry *proc_net_create(const char *name,
				       mode_t mode, get_info_t *get_info);
void proc_net_remove(const char *name);
unsigned short ip_fast_csum(unsigned char *iph, unsigned int ihl);
unsigned long get_millis();

void platform_init();
void platform_fini();

char *inet_ntoa(struct in_addr in);
int inet_aton(char *cp, struct in_addr *ap);

// ######################################################################
#else // Linux (Intel and ARM)

#include "platform_lnx.h"

#endif // Linux


// ######################################################################
// Common stuff needed by both versions

#define htonsel(s) cpu_to_be64(s)
#define ntohsel(s) be64_to_cpu(s)

#ifndef MAX
#  define MAX(a,b) ((a) < (b) ? (b) : (a))
#  define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#endif //NSPORT

#endif // PLATFORM_WIN_H
