/*ns version of netbox*/

int NETBOX_CLASS netbox_l2send(int ifnr, struct sockaddr *dest, sapfsel_t sel,
		    void *data, int len){

  Mac *mac_=selnet_->get_mac();
  
  NsObject *ifq=selnet_->get_ifq();
  double delay_=selnet_->get_delay();
  nsaddr_t my_IP=selnet_->get_IP();
  int dest_addr;
  if (!dest)
    dest_addr = MAC_BROADCAST;
  else
    ethtoint(dest->sa_data,&dest_addr);
  
  
  
  DEBUG(10, printk("netbox_l2send %d 0x%016Lx (%d bytes)\n",
		   dest_addr, ntohsel(sel), len));
  
  /*prepare a packet*/
  Packet *p;
  hdr_cmn *ch;
  //if len=-1 then there is a packet inside
  if(len!=-1)
    {       
      p = Packet::alloc();
      ch = HDR_CMN(p);
      ch->ptype()=PT_XRP;
    }
  else
    {
      p=(Packet*)data;
      ch = HDR_CMN(p);
    }
  char *mh=(char*)HDR_MAC(p);
  hdr_ll	*lh = HDR_LL(p);
  hdr_rtp *rh = HDR_RTP(p);
  hdr_tcp *th = HDR_TCP(p);
  hdr_selnet *sh = HDR_SELNET(p);
  hdr_xrp *xh = HDR_XRP(p);
  
  
  mac_->hdr_dst(mh, dest_addr);
  mac_->hdr_src(mh, mac_->addr());
  mac_->hdr_type(mh, SAPF_ETHERTYPE);
  
  lh->lltype() = LL_DATA;
  
  ch->direction() = hdr_cmn::DOWN; // send this pkt down
  
  sh->sel=sel; 
  
  
  
  if(len!=-1)
    {      	  
      memcpy(((char*)xh->data),(char*)data, len);
      ch->size() = len+sizeof(sapfsel_t);//Len is data. add length of selector
    }
  else
    {
      if((int)*xh->data ==-1)//packet generated at this node by an agent
	*(xh->data)=(char)ch->ptype();

      //FINAL type of packet 
      ch->ptype() = PT_SELNET;
      ch->size() += sizeof(sapfsel_t);//Len is data. add length of selector  
    }
  //send the packet
  ifq->recv(p,(Handler*) 0);
  return len;

}


int NETBOX_CLASS netbox_arpset(struct in_addr addr, struct sockaddr *hw_addr, struct net_device *dev){


  Mac *mac_=selnet_->get_mac();
  LL *ll_=selnet_->get_ll();
  NsObject *ifq=selnet_->get_ifq();
  
   //create a new packet
  Packet *p = Packet::alloc();

  //access all the headers needed for ARP
  hdr_cmn *ch = HDR_CMN(p);
  char	*mh = (char*)HDR_MAC(p);
  hdr_ll  *lh = HDR_LL(p);
  hdr_arp *ah = HDR_ARP(p);

  //prepare the data
  nsaddr_t my_IP=selnet_->get_IP();
  nsaddr_t src_IP=addr.s_addr;

  int fake_mac; //Fake MAC source
  ethtoint((char*)hw_addr->sa_data,&fake_mac);
  fake_mac=ntohl(fake_mac);
  int my_mac=mac_->addr();

  //prepare a fake arp reply


		ch->size() = ARP_HDR_LEN;
		ch->error() = 0;
		ch->direction() = hdr_cmn::UP; // send this pkt UP to LL
		ch->ptype() = PT_ARP;

		mac_->hdr_dst(mh, my_mac);
		mac_->hdr_src(mh, fake_mac);
		mac_->hdr_type(mh, ETHERTYPE_ARP);

		lh->seqno() = 0;
		lh->lltype() = LL_DATA;

		ah->arp_op  = ARPOP_REPLY;
		ah->arp_tha = my_mac;
		ah->arp_sha = fake_mac;

		
		ah->arp_spa = src_IP;
		ah->arp_tpa = my_IP;



		//SEND this packet UP
		ll_->recv(p,0);
}


void NETBOX_CLASS netbox_deltimer(netbox_timer_t *t){
  //don't need it just to make real code happy
}

int NETBOX_CLASS netbox_hasDGW(void){
  //don't need it just to make real code happy
}


void NETBOX_CLASS netbox_newtimer(netbox_timer_t *t,
				  fct_t fct, void *data, unsigned long delta_usec){
  double now=Scheduler::instance().clock();
  double expires =1.0*delta_usec /1000000;// expires now in seconds as needed
  
  DEBUG(10, printk("\n\nNETBOX::newtimer: expires after %f, now=%f \n\n",expires, now));
  
  t->init(expires,fct, data);
  t->resched(expires); 
  t->init(0,fct, data);

}
int NETBOX_CLASS  netbox_init(){
  return 0;
  //do not use
}
int NETBOX_CLASS netbox_addL2if(int ifnr, int handle){
  return 0;
  //do not use
}
int NETBOX_CLASS  netbox_dropL2if(int ifnr){
  return 0;
  //do not use
}
void NETBOX_CLASS netbox_fini(){
  //do not use

}
int NETBOX_CLASS netbox_getL2addr(int ifnr, struct sockaddr *sa){
  Mac *mac_=selnet_->get_mac();
  int addr=mac_->addr();
  memcpy(sa->sa_data,(char*)&addr,ETH_ALEN);
  DEBUG(10, printk("getL2addr: from MAC %d written: %d\n", mac_->addr(), (int)sa->sa_data));

}


int NETBOX_CLASS netbox_getHID128(unsigned char *buf){
  //reseved for future work. will not be present in NS 
}


int NETBOX_CLASS netbox_getL3addr(struct in_addr *name, struct net_device *dev){
  nsaddr_t ip_=selnet_->get_IP();
  name->s_addr=(int)ip_;
  DEBUG(10, printk("getL3addr: %d\n", name->s_addr));


}


int NETBOX_CLASS netbox_arpdelete(struct in_addr addr, struct net_device *dev){

  ARPTable* arptable_=selnet_->get_arptable();
  ARPEntry *arp_entry_=arptable_->arplookup((nsaddr_t)addr.s_addr);
   DEBUG(10, printk("NETBOX::arpdelete\n"));
 
  arp_entry_->entrydown();
}


int NETBOX_CLASS netbox_l3deliver(int ifnr, void *data, int len){
  Packet *p;
  if(len!=-1)
    {
      printf("ERROR:: NETBOX trrying to deliver non-ip packet toi IP level\n");  
      exit(0);
    }
  p=(Packet *)data;
  
  hdr_cmn *ch = HDR_CMN(p);
  hdr_xrp *xh=HDR_XRP(p);
  hdr_rtp *rh=HDR_RTP(p);
  hdr_tcp *th=HDR_TCP(p);
  
  ch->ptype()=(packet_t)*(xh->data);
  //already  subtracted at macinput ch->size() -= sizeof(sapfsel_t);
  LL *ll_=selnet_->get_ll(); 
  ll_->recv(p,0);

}
