/* selnet.cc
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Lunar port to ns-2 
 *
 * Copyright (C) 2001-2003 <christian.tschudin@unibas.ch>
 * Ported by Evgeny Osipov <evgeny.osipov@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */
#ifndef NS_PORT
#define NS_PORT
#endif


#include "../include/platform.h"

#include <errno.h>



#include "delay.h"
#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "rtp.h"
#include "tcp.h"
#include "mac-802_11.h"
#include "selnet.h"
#include "sapf_ns.h"
#include "netbox_ns.h"
#include "topography.h"
#include "cmu-trace.h"
#include "mobilenode.h"
#include "tools/random.h"

#include <address.h>

static class SELNETClass : public TclClass {
public:
        SELNETClass() : TclClass("SELNET") {}
        TclObject* create(int, const char*const* argv) {
                return (new SELNET(argv[4], argv[5], argv[6], argv[7], argv[8]));
        }
} class_selnet;

int hdr_selnet::offset_;

static class SELNETHeaderClass : public PacketHeaderClass {
public:
        SELNETHeaderClass() : PacketHeaderClass("PacketHeader/SELNET",
                                            sizeof(hdr_selnet)) { 
		bind_offset(&hdr_selnet::offset_);
	}
} class_hdr_selnet;

int hdr_xrp::offset_;

static class XRPHeaderClass : public PacketHeaderClass {
public:
        XRPHeaderClass() : PacketHeaderClass("PacketHeader/XRP",
                                            sizeof(hdr_xrp)) { 
		bind_offset(&hdr_xrp::offset_);
	} 
} class_hdr_xrp;




unsigned int Random_byte=155;

SELNET::SELNET(const char *tclnode, const char *tclmac, const char *tclll,const char *arptable,const char *ifq ) : LinkDelay()  
{ 

  node_ = (MobileNode*) TclObject::lookup(tclnode);
  assert(node_);
  My_IP=node_->address();
  Node *tmp=(Node*)node_;

  mac_ = (Mac*) TclObject::lookup(tclmac);
  assert(mac_);
  
  ll_ = (LL*) TclObject::lookup(tclll);
  assert(ll_);
  
  arptable_ = (ARPTable*) TclObject::lookup(arptable);
  assert(arptable_);
  
  ifq_ = (NsObject*) TclObject::lookup(ifq);
  assert(arptable_);
  
	
  /*INITIALIZE MY FIELDS FIRST*/
  // XRP_CTYPE_UNKNOWN:
  xrp_param_len[0]=  0;
  // XRP_CTYPE_SEL:
  xrp_param_len[1]=   sizeof(struct xrp_head_s) + sizeof(sapfsel_t);
  // XRP_CTYPE_IPv4
  xrp_param_len[2]= sizeof(struct xrp_head_s) + sizeof(uint32_t);
  // XRP_CTYPE_IPv6
  xrp_param_len[3]= sizeof(struct xrp_head_s) + 16;
  // XRP_CTYPE_SEL_ETH
  xrp_param_len[4]= sizeof(struct xrp_head_s) + sizeof(sapfsel_t) + ETH_ALEN;
  // XRP_CTYPE_SEL_UDP
  xrp_param_len[5]= sizeof(struct xrp_head_s) + sizeof(sapfsel_t) + sizeof(uint32_t)
    + sizeof(short);
  // XRP_CTYPE_HID128
  xrp_param_len[XRP_PARTYPE_COUNT]=  sizeof(struct xrp_head_s) + 16;
  
  
  stats=new struct net_device_stats;
  /*create sapf object here*/
  sapf_=new SAPF(this);
  sapf_->sapf_init();
  //Initialize FIB array
  for(int i=0;i<SAPF_MAX_FIB_TABLES;i++)
    {
      sapf_->fib[i]=new struct fibtable_s; 
      memset(sapf_->fib[i], 0, sizeof(struct fibtable_s));
    }
  
  



  bind("conf_min_ring", &conf_min_ring);
  bind("conf_max_ring", &conf_max_ring);
  bind("conf_timeout_per_ring", &conf_timeout_per_ring);
  bind("conf_retry_wait", &conf_retry_wait);
  bind("conf_max_retries", &conf_max_retries);
  bind("conf_fri", &conf_fri);
  bind("conf_fri_jitter", &conf_fri_jitter);
  bind("debuglevel", &debuglevel);
  
  forwarding="yes";
  
  /*create netbox object here*/
  netbox_=new NETBOX(this);
  for(int i=0;i<LUNAR_MAX_PEERS;i++)
    { 
      peers[i].timer.set_callback(this, debuglevel);
    }
  
  lunardev=new struct net_device;
  
  lunardev->ifindex=0;
  lunar_init(); 
  
  DEBUG(10, printk("MYIP=%d\n",My_IP));
  DEBUG(10, printk("initialized \n")); 
}

int
SELNET::command(int argc, const char*const* argv)
{
	if (argc == 2 && strcasecmp(argv[1], "reset") == 0) {
	}
	return LinkDelay::command(argc, argv);
}

void SELNET::handle(Event* e)
{
  recv((Packet *)e,0);
}
#include "../src/lunar.c"

#include "../src/xrp.c"

#include "netbox_ns.cc"

void SELNET::recv(Packet* p, Handler* /*h*/)
{	
  hdr_cmn *ch = HDR_CMN(p);
  DEBUG(10, printk("SELNET::recv:PACKET - Direction= %d (DOwN=%d UP=%d default=-1)\n",ch->direction(),hdr_cmn::DOWN,hdr_cmn::UP));

  // If direction = UP, then pass it up the stack
  // Otherwise, set direction to DOWN and pass it down the stack
  if(ch->direction() == hdr_cmn::UP) {
    if(ch->ptype_ == PT_SELNET || ch->ptype_ == PT_XRP)
      macinput(p);
    else //end Lunars'extensions //this should never happen in Lunar
      // ALL PACKETS ARE SELNET
      ll_ ? ll_->recv(p,0) : drop(p); // send to ll it will handle it.
    return;
  }
  
  ch->direction() = hdr_cmn::DOWN;
  ipinput(p);
}


 
int
SELNET::ipinput(Packet *p)
{

  /*THIS METHOD IS AN ANALOGUE TO klunar_start_xmit()*/
  Scheduler& s = Scheduler::instance();

  hdr_cmn *ch = HDR_CMN(p);
  char *mh=(char*)HDR_MAC(p);
  hdr_ll	*lh = HDR_LL(p);
  hdr_arp	*ah = HDR_ARP(p);
  hdr_selnet *sh = HDR_SELNET(p);
  hdr_xrp *xh = HDR_XRP(p);
  hdr_ip *ih = HDR_IP(p);

  struct hdr_mac802_11 *mh_802_11=(struct hdr_mac802_11 *)mh;
  
  nsaddr_t dst_= (nsaddr_t)Address::instance().get_nodeaddr(ih->daddr());
  
  nsaddr_t src_ = (nsaddr_t)Address::instance().get_nodeaddr(ih->saddr());
   //make real code happy
  struct in_addr addr;
  struct net_device *dev=basedev;

  int len = ch->size();
  u_int16_t typeMAC= GET2BYTE(mh_802_11->dh_body);//this macro is from marshal.h included from mac-802_11.h

  /*THIS IS FROM klunar_start_xmit*/
  sapfsel_t sel;
  //sapfpkt pkt;// = new sapfpkt;
  nsaddr_t targetIP;
  int res;
  DEBUG(10, printk("klunar start_xmit (%d bytes)\n", len));

  switch (typeMAC) 
    {
    case ETHERTYPE_ARP:
      DEBUG(10, printk("klunar - arp detected\n"));
      targetIP=ah->arp_tpa;
      addr.s_addr=targetIP;
      res=lunar_do_arp(addr);
      if(res==-1)
	{
	  // DEBUG(10, printk("\n\n\n arp DROPED\n\n\n\n"));
	  //q_handler->handle((Event*)0);//keep the queue active
	};
      kfree_skb(p);//new packet will be created
      break;
    case ETHERTYPE_IP:
      if (stats) {
	stats->tx_packets++;
	stats->tx_bytes += len;
      }
      {
	int i;
	DEBUG(10, printk("IP data (%d):", len));
	DEBUG(10, printk("\n"));
      }
      *(xh->data)=(char)-1;//this is to indicate that packet is originated at this node
      pkt_ip.data = (char*)p;// put a pointer to real packet here
      pkt_ip.len = -1;//this is to indicate that there is a packet inside //skb->len - dev->hard_header_len;
      pkt_ip.ifnr = -1;
      int MAC_DA_INT;
      
      if (ETHER_ADDR(mh_802_11->dh_da)==MAC_BROADCAST)
	{
	  // broadcast
	  if (bcast_selector)
	    sapf_->sapf_demux(mysapfhandle, bcast_selector, &pkt_ip);
	  else
	    lunar_do_bc_setup(&bcast_selector, &pkt_ip);
	}     
      else { // unicast
	sel = 0;
	MAC_DA_INT=htonl(GET4BYTE(mh_802_11->dh_da));
	memcpy(((char*)&sel) + sizeof(sel) - ETH_ALEN,
	       ((char*)&MAC_DA_INT),ETH_ALEN);
	sel|=htonsel(SELCLASS_LOCAL);

	sapf_->sapf_demux(mysapfhandle, sel, &pkt_ip);
	}
      break;
    default:
      kfree_skb(p);
      break;
    }
  
  return 0;
  
  

} 

void
SELNET::macinput(Packet *p)
{
  //this function is an analog to l2recv()

   Scheduler& s = Scheduler::instance();
   double now=s.clock();

  hdr_cmn *ch = HDR_CMN(p);
  hdr_ip *ih = HDR_IP(p);
  hdr_selnet *sh = HDR_SELNET(p);
  hdr_xrp *xh = HDR_XRP(p);
  hdr_tcp *th = HDR_TCP(p);
  sapfsel_t sel= sh->sel;
  
  //sapfpkt *pkt=new sapfpkt;
  if(ch->ptype()==PT_XRP)
    {

     	memcpy(xrpdata, xh->data,2000);
	//xrpdata=xh->data;
      pkt.data = (char*)xrpdata;

      struct xrp_head_s *x = (struct xrp_head_s*) pkt.data;
      pkt.len = ch->size()-sizeof(sapfsel_t); //subtract the selector's size
      pkt.ifnr = -1;

      kfree_skb(p);
    }else //this is selnet packet but not RREQ/RREP
      {
	ch->size()-=sizeof(sapfsel_t);
	pkt.data = (char*)p;
	pkt.len=-1;
	pkt.ifnr=-1;
      }

  DEBUG(10, printk("SELNET::macinput: Selector %016Lx\n", ntohsel(sel)));

  sapf_->sapf_demux(mysapfhandle,sel,&pkt);
 
}

#include "../src/lunar_init.c"


/*=======================================================================
SAPF staff: Constructor, TIMERŝ expire and redefined functtions
==========================================================================*/
//SAPF constructor
SAPF::SAPF(SELNET *a):sapf_gc_timer(this)
{
  selnet_=a;
  debuglevel=selnet_->debuglevel;
}

SAPFTimer::SAPFTimer(SAPF *a):TimerHandler()
{
  a_=a; 
  debuglevel=a_->debuglevel;
}

//NETBOX constructor
NETBOX::NETBOX(SELNET *a)
{
  selnet_=a;
  debuglevel=selnet_->debuglevel;
}

void SAPF::add_timer(SAPFTimer *a)
{
  double timeout = a->expires;
  double now=Scheduler::instance().clock();
  a->resched(timeout); 
  a->expires=0.0;
}


void SAPF::del_timer(SAPFTimer *a)
{
  
  a->cancel();
 
}
void SAPF::get_random_bytes (sapfsel_t *sel, int nbytes)
{
  unsigned char *cp = (unsigned char *) sel;
  if (nbytes > 2)
    nbytes = 2;
  while (nbytes-- > 0) {
    *cp++ = Random::integer(256);
  }
}
void SAPFTimer::expire(Event *e)
{
  debuglevel=a_->debuglevel;
  double now=Scheduler::instance().clock();
  DEBUG(10, printk("sapf_gc:: now=%f\n", now));
  a_->sapf_gc(0);
}
void NETBOXTimer::expire(Event *e)
{
  double now=Scheduler::instance().clock();
  DEBUG(10, printk("\n\nNETBOX::timer:expire: %f \n", now));
  
  (a_->*fct)(data);
}

#include "../src/sapf.c"
 
  
 
 
 
 
