//Header for selnet.cc
#ifndef __selnet_h__
#define __selnet_h__

#include "../include/platform.h"

#include "scheduler.h"
#include "delay.h"
#include "timer-handler.h"

class LinkDelay;
class LL;
class Mac;
class SELNET;
class SAPF;
class NETBOX;

typedef void (SELNET::*fct_t)(void *);

class NETBOXTimer: public TimerHandler {
 public:
  NETBOXTimer(): TimerHandler() {a_=0;}
  void set_callback(SELNET *a, int debug){a_=a; debuglevel=debug;}
  void init(double expires_,  fct_t fct_,void *data_ ){expires=expires_;data=data_;fct=fct_;}
   int debuglevel;
  protected:
  virtual void expire (Event *e);
  SELNET *a_;
  double expires;
  void *data;
  fct_t fct;
};


#include "../include/sapf.h"
#include "../include/netbox.h"
#include "../include/lunar.h"
#include "../include/xrp.h"

//SELNET and XRP headers

struct hdr_selnet { 
  sapfsel_t sel;
  // Header access methods
  static int offset_; // required by PacketHeaderManager
  inline static int& offset() { return offset_; }
  inline static hdr_selnet* access(const Packet* p) {
    return (hdr_selnet*) p->access(offset_);
  }
};

struct hdr_xrp { 
  char data[2000];// change if maximum XRP msg size will change
  // Header access methods
  static int offset_; // required by PacketHeaderManager
  inline static int& offset() { return offset_; }
  static hdr_xrp* access(const Packet* p) {
    return (hdr_xrp*) p->access(offset_);
  }
};


class SELNET : public LinkDelay {
public:
  
//NS staff

	SELNET(const char *tclnode, const char *tclmac, const char *tclll, const char *arptable,const char *ifq);

        int     command(int argc, const char*const* argv);
	void recv(Packet* p, Handler* qh);
	void handle(Event* e);
	//THESE TWO METHODS ARE INTERFACES for LL to ACCESS XRP
	//This is done in order do not change LL
	int     ipinput(Packet *p);
	void    macinput(Packet *p);
	int debug_;

	/*Methods to retrieve NODE MAC and LL instances*/
	MobileNode* get_node() {return node_;}
	Mac* get_mac(){return mac_;}
	LL* get_ll(){return ll_;}
	NsObject* get_ifq(){return ifq_;}
	ARPTable* get_arptable(){return arptable_;}
	double get_delay(){return delay_;}
	nsaddr_t get_IP(){return My_IP;}

// Lunar staff
#include "../include/lunar_decl.h"

// XRP staff
#include "../include/xrp_decl.h" 

	sapfpkt pkt_ip;
	sapfpkt pkt;
	char xrpdata[2000];


private:
	//NS staff
	MobileNode	*node_;
	Mac		*mac_;
	LL		*ll_;
	NsObject       *ifq_;

	//LUNAR staff
	inline int initialized() { return node_ && mac_; }
	int lunar_init(void);
	void lunar_exit();
	struct peer_s peers[LUNAR_MAX_PEERS];
	struct net_device_stats *stats;
	
	//porting staff
	SAPF *sapf_;
	NETBOX *netbox_;

	ARPTable *arptable_;
        NETBOXTimer *ntimer_;


	//auxillary fields
	nsaddr_t My_IP;
};

#endif /* __selnet_h__ */
